#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import shutil
import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory
from .ara import _validate_ara_rtl

logger = get_logger()

Tasks = get_task_factory()


def _validate_spike(context: MlonMcuContext, params=None):
    if not context.environment.has_target("spike") and not _validate_ara_rtl(context, params=params):
        return False
    user_vars = context.environment.vars
    if "spike.pk" not in user_vars:  # TODO: also check command line flags?
        assert "spikepk" in context.environment.repos, "Undefined repository: 'spikepk'"
    if "spike.exe" not in user_vars:  # TODO: also check command line flags?
        assert "spike" in context.environment.repos, "Undefined repository: 'spike'"
    return True


def _validate_spikepk(context: MlonMcuContext, params=None):
    if not _validate_spike(context, params=params):
        return False
    user_vars = context.environment.vars
    # multilib = user_vars.get("riscv_gcc.multilib", False)
    supported_archs = user_vars.get("riscv_gcc.supported_archs", [])
    if params:
        arch = params["arch"]
        # if arch != "rv32gc":
        if arch != "default":
            if arch not in supported_archs:
                return False
    return True


def _validate_spike_clean(context: MlonMcuContext, params={}):
    if not _validate_spike(context, params=params):
        return False
    user_vars = context.environment.vars
    keep_build_dir = user_vars.get("spike.keep_build_dir", True)
    return not keep_build_dir


@Tasks.provides(["spikepk.src_dir"])
@Tasks.validate(_validate_spikepk)
@Tasks.register(category=TaskType.TARGET)
def clone_spike_pk(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the spike proxy kernel."""
    spikepkName = utils.makeDirName("spikepk")
    spikepkSrcDir = context.environment.paths["deps"].path / "src" / spikepkName
    user_vars = context.environment.vars
    if "spike.pk" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(spikepkSrcDir):
        spikepkRepo = context.environment.repos["spikepk"]
        utils.clone_wrapper(spikepkRepo, spikepkSrcDir, refresh=rebuild)
    context.cache["spikepk.src_dir"] = spikepkSrcDir


@Tasks.needs(["spikepk.src_dir", "riscv_gcc.install_dir", "riscv_gcc.name"])
@Tasks.provides(["spikepk.build_dir", "spikepk.install_dir", "spike.pk"])
# TODO: allow arch,abi
@Tasks.param("arch", ["default"])  # ["rv32gc", "rv64gc", "rv32im", "rv64im"]
@Tasks.validate(_validate_spikepk)
@Tasks.register(category=TaskType.TARGET)
def build_spike_pk(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build Spike proxy kernel."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    if "spike.pk" in user_vars:  # TODO: also check command line flags?
        return False
    spikepkName = utils.makeDirName("spikepk")
    spikepkSrcDir = context.cache["spikepk.src_dir"]
    spikepkBuildDir = context.environment.paths["deps"].path / "build" / spikepkName
    spikepkInstallDir = context.environment.paths["deps"].path / "install" / spikepkName
    # default_arch = "rv32gc"
    arch = params.get("arch", "rv32gc")
    if arch == "default":
        arch = user_vars.get("spikepk.default_arch", "rv32imafdc_zifencei_zicsr")
        abi = user_vars.get("spikepk.default_abi", "ilp32d")
    else:
        abi = "lp64" if "rv64" in arch else "ilp32"
    # spikepkBin = spikepkInstallDir / f"pk_{arch}_{abi}"
    spikepkDefaultBin = spikepkInstallDir / "pk"
    # if rebuild or not (utils.is_populated(spikepkBuildDir) and spikepkBin.is_file()):
    if rebuild or not (utils.is_populated(spikepkBuildDir) and spikepkDefaultBin.is_file()):
        # No need to build a vext and non-vext variant?
        utils.mkdirs(spikepkBuildDir)
        gccName = context.cache["riscv_gcc.name"]
        # assert gccName == "riscv32-unknown-elf", "Spike PK requires a non-multilib toolchain!"
        if "riscv_gcc.install_dir" in user_vars:
            riscv_gcc = user_vars["riscv_gcc.install_dir"]
        else:
            riscv_gcc = context.cache["riscv_gcc.install_dir"]
        spikepkArgs = []
        spikepkArgs.append(f"--with-arch={arch}")
        spikepkArgs.append(f"--with-abi={abi}")
        spikepkArgs.append("--prefix=" + str(riscv_gcc))
        spikepkArgs.append("--host=" + gccName)
        env = os.environ.copy()
        env["PATH"] = str(Path(riscv_gcc) / "bin") + ":" + env["PATH"]
        utils.exec_getout(
            str(spikepkSrcDir / "configure"),
            *spikepkArgs,
            cwd=spikepkBuildDir,
            env=env,
            live=False,
        )
        utils.make(cwd=spikepkBuildDir, threads=threads, live=verbose, env=env)
        # utils.make(target="install", cwd=spikepkBuildDir, live=verbose, env=env)
        utils.mkdirs(spikepkInstallDir)
        # utils.move(spikepkBuildDir / "pk", spikepkBin)
        # if arch == default_arch:
        utils.copy(spikepkBuildDir / "pk", spikepkDefaultBin)
    context.cache["spikepk.build_dir"] = spikepkBuildDir
    context.cache["spikepk.install_dir"] = spikepkInstallDir
    # if arch == default_arch:
    context.cache["spike.pk"] = spikepkDefaultBin
    context.export_paths.add(spikepkInstallDir)


@Tasks.provides(["spike.src_dir"])
@Tasks.validate(_validate_spike)
@Tasks.register(category=TaskType.TARGET)
def clone_spike(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the spike simulator."""
    spikeName = utils.makeDirName("spike")
    spikeSrcDir = context.environment.paths["deps"].path / "src" / spikeName
    user_vars = context.environment.vars
    if "spike.exe" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(spikeSrcDir):
        spikeRepo = context.environment.repos["spike"]
        utils.clone_wrapper(spikeRepo, spikeSrcDir, refresh=rebuild)
    context.cache["spike.src_dir"] = spikeSrcDir


@Tasks.needs(["spike.src_dir", "riscv_gcc.install_dir", "riscv_gcc.name"])
@Tasks.provides(["spike.build_dir", "spike.exe"])
@Tasks.validate(_validate_spike)
@Tasks.register(category=TaskType.TARGET)
def build_spike(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build Spike simulator."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    if "spike.exe" in user_vars:  # TODO: also check command line flags?
        return False
    spikeName = utils.makeDirName("spike")
    spikeSrcDir = context.cache["spike.src_dir"]
    spikeBuildDir = context.environment.paths["deps"].path / "build" / spikeName
    spikeInstallDir = context.environment.paths["deps"].path / "install" / spikeName
    spikeExe = spikeInstallDir / "spike"
    user_vars = context.environment.vars
    if "spike.exe" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not (utils.is_populated(spikeBuildDir) and spikeExe.is_file()):
        # No need to build a vext and non-vext variant?
        utils.mkdirs(spikeBuildDir)
        spikeArgs = []
        # spikeArgs.append("--prefix=" + str(context.cache["riscv_gcc.install_dir"]))
        spikeArgs.append("--prefix=" + str(spikeInstallDir))
        spikeArgs.append("--enable-misaligned")
        utils.exec_getout(
            str(Path(spikeSrcDir) / "configure"),
            *spikeArgs,
            cwd=spikeBuildDir,
            live=False,
        )
        utils.make(cwd=spikeBuildDir, threads=threads, live=verbose)
        utils.make("install", cwd=spikeBuildDir, threads=threads, live=verbose)
        utils.mkdirs(spikeInstallDir)
        utils.move(spikeBuildDir / "spike", spikeExe)
    context.cache["spike.build_dir"] = spikeBuildDir
    context.cache["spike.install_dir"] = spikeInstallDir
    context.cache["spike.exe"] = spikeExe
    context.export_paths.add(spikeInstallDir)


@Tasks.needs(["spike.exe", "spike.build_dir"])  # TODO: make sure spike.exe has beeen copies before
@Tasks.removes(["spike.build_dir"])  # TODO: implement
@Tasks.validate(_validate_spike_clean)
@Tasks.register(category=TaskType.TARGET)
def clean_spike(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Cleanup Spike build dir."""
    spikeBuildDir = context.cache["spike.build_dir"]
    shutil.rmtree(spikeBuildDir)
    del context.cache["spike.build_dir"]


def _validate_microtvm_spike(context: MlonMcuContext, params=None):
    return context.environment.has_target("microtvm_spike")


@Tasks.provides(["microtvm_spike.src_dir", "microtvm_spike.template"])
@Tasks.validate(_validate_microtvm_spike)
@Tasks.register(category=TaskType.TARGET)
def clone_microtvm_spike(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the microtvm-spike-template repository."""
    name = utils.makeDirName("microtvm_spike")
    srcDir = context.environment.paths["deps"].path / "src" / name
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["microtvm_spike"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    context.cache["microtvm_spike.src_dir"] = srcDir
    context.cache["microtvm_spike.template"] = srcDir / "template_project"



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from mlonmcu.target.target import Target

from mlonmcu.logging import get_logger

from .microtvm_template_target import TemplateMicroTvmPlatformTarget

logger = get_logger()


class EspidfMicroTvmPlatformTarget(TemplateMicroTvmPlatformTarget):
    DEFAULTS = {
        **Target.DEFAULTS,
        "project_type": "host_driven",
        "verbose": False,
        "board": None,
        "port": None,
        "baud": None,
    }
    REQUIRED = Target.REQUIRED | {"microtvm_espidf.template", "espidf.src_dir", "espidf.install_dir"}

    def __init__(self, name=None, features=None, config=None):
        super().__init__(name=name, features=features, config=config)
        self.template_path = self.microtvm_espidf_template
        # TODO: interate into TVM build config
        self.option_names = [
            "project_type",
            "verbose",
            "idf_path",
            "idf_tools_path",
            "idf_target",
            "idf_serial_port",
            "idf_serial_baud",
            "warning_as_error",
            "compile_definitions",
            "extra_files_tar",
        ]

    @property
    def microtvm_espidf_template(self):
        return Path(self.config["microtvm_espidf.template"])

    @property
    def esp_idf_src_dir(self):
        return Path(self.config["espidf.src_dir"])

    @property
    def esp_idf_install_dir(self):
        return Path(self.config["espidf.install_dir"])

    @property
    def board(self):
        return self.config["board"]

    @property
    def port(self):
        return self.config["port"]

    @property
    def baud(self):
        return self.config["baud"]

    def get_project_options(self):
        ret = super().get_project_options()
        ret.update({"idf_path": self.esp_idf_src_dir})
        ret.update({"idf_tools_path": self.esp_idf_install_dir})
        assert self.board, "microtvm_espidf.board has do be defined"
        ret.update({"idf_target": self.board})
        if self.port:
            ret.update({"idf_serial_port": self.port})
        if self.baud:
            ret.update({"idf_serial_baud": self.baud})
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for the artifact submodule."""

from mlonmcu.artifact import Artifact, ArtifactFormat, lookup_artifacts


def test_lookup_artifacts():
    first = Artifact("foo.json", content="{}", fmt=ArtifactFormat.SOURCE, flags={"metadata"})
    second = Artifact("bar.txt", content="bar", fmt=ArtifactFormat.TEXT)
    third = Artifact("hello.elf", raw=b"", fmt=ArtifactFormat.RAW, flags={"sw", "elf", "test"})
    fourth = Artifact("world.txt", content="", fmt=ArtifactFormat.TEXT, flags={"test"})
    artifacts = [first, second, third, fourth]
    assert lookup_artifacts(artifacts) == artifacts
    assert lookup_artifacts(artifacts, first_only=True) == [first]
    assert lookup_artifacts(artifacts, name="foo.txt") == []
    assert lookup_artifacts(artifacts, name="bar.txt") == [second]
    assert lookup_artifacts(artifacts, fmt=ArtifactFormat.TEXT) == [second, fourth]
    assert lookup_artifacts(artifacts, fmt=ArtifactFormat.TEXT, first_only=True) == [second]
    assert lookup_artifacts(artifacts, fmt=ArtifactFormat.RAW) == [third]
    assert lookup_artifacts(artifacts, flags={"test"}) == [third, fourth]
    assert lookup_artifacts(artifacts, flags={"test", "sw"}) == [third]



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC

from mlonmcu.feature.type import FeatureType
from mlonmcu.config import filter_config
from mlonmcu.feature.features import get_matching_features


class Framework(ABC):
    registry = {}

    name = None

    FEATURES = set()
    DEFAULTS = {}
    REQUIRED = {"tf.src_dir"}
    OPTIONAL = set()

    def __init__(self, features=None, config=None, backends={}):
        self.config = config if config else {}
        self.features = self.process_features(features)
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)

    def process_features(self, features):
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.FRAMEWORK)
        for feature in features:
            assert feature.name in self.FEATURES, f"Incompatible feature: {feature.name}"
            feature.used = True
            feature.add_framework_config(self.name, self.config)
        return features

    def remove_config_prefix(self, config):
        def helper(key):
            return key.split(f"{self.name}.")[-1]

        return {helper(key): value for key, value in config if f"{self.name}." in key}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        assert isinstance(cls.name, str)
        cls.registry[cls.name] = cls

    def get_platform_config(self, platform):
        return {}

    def add_platform_config(self, platform, config):
        config.update(self.get_platform_config(platform))

    def get_platform_defs(self, platform):
        if platform == "espidf":
            framework_upper = self.name.upper()
            return {f"MLONMCU_FRAMEWORK_{framework_upper}": True}
        else:
            return {"MLONMCU_FRAMEWORK": self.name}

    def add_platform_defs(self, platform, defs):
        defs.update(self.get_platform_defs(platform))



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for cleaning up the current environment."""
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import setup

from mlonmcu.cli.common import (
    add_common_options,
    add_context_options,
)


def get_parser(subparsers):
    """ "Define and return a subparser for the cleanup subcommand."""
    parser = subparsers.add_parser("cleanup", description="Cleanup ML on MCU environment.")
    parser.set_defaults(func=handle)
    add_common_options(parser)
    add_context_options(parser)
    parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Do not ask before removing disk contents (DANGEROUS)",
    )
    parser.add_argument(
        "-k",
        "--keep",
        metavar="KEEP",
        type=int,
        default=10,
        help="Remove everything except the latest KEEP sessions (default: %(default)s)",
    )
    parser.add_argument(
        "--deps",
        default=False,
        action="store_true",
        help="Also delete all dependencies from the environment.",
    )
    parser.add_argument(
        "--cache",
        default=False,
        action="store_true",
        help="Clear the environments dependency cache.",
    )
    return parser


def handle(args):
    with MlonMcuContext(path=args.home, deps_lock="write") as context:
        interactive = not args.force
        keep = args.keep
        context.cleanup_sessions(keep=keep, interactive=interactive)
        installer = setup.Setup(context=context)
        if args.deps:
            # This will also remove the cache file
            installer.clean_dependencies(interactive=interactive)
        elif args.cache:
            installer.clean_cache(interactive=interactive)



import pytest

# from mlonmcu.artifact import ArtifactFormat
from mlonmcu.artifact import lookup_artifacts

from mlonmcu.testing.helpers import (
    DEFAULT_BACKENDS,
    DEFAULT_TFLITE_MODELS,
    DEFAULT_MLIF_TARGETS,
    DEFAULT_ESPIDF_TARGETS,
    _test_run_platform,
    _test_compile_platform,
)


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("backend_name", DEFAULT_BACKENDS)
@pytest.mark.parametrize("target_name", DEFAULT_MLIF_TARGETS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize("config", [{}])
def test_target_mlif(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    _, artifacts = _test_run_platform(
        "mlif", backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )

    assert len(lookup_artifacts(artifacts, name="generic_mlonmcu")) == 1
    assert len(lookup_artifacts(artifacts, name="mlif_out.log")) == 1
    # TODO: check artifacts


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("backend_name", DEFAULT_BACKENDS)
@pytest.mark.parametrize("target_name", DEFAULT_ESPIDF_TARGETS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config", [{"espidf.wait_for_user": False, "tflmi.arena_size": 2**17, "tvmaot.arena_size": 2**17}]
)
def test_platform_espidf(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    _, artifacts = _test_compile_platform(
        "espidf", backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )

    assert len(lookup_artifacts(artifacts, name="generic_mlonmcu")) == 1
    assert len(lookup_artifacts(artifacts, name="espidf_out.log")) == 1
    # TODO: check artifacts


@pytest.mark.slow
@pytest.mark.hardware
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("backend_name", DEFAULT_BACKENDS)
@pytest.mark.parametrize("target_name", ["esp32c3"])
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config",
    [
        {
            "espidf.wait_for_user": False,
            "espidf.use_idf_monitor": False,
            "espidf.port": "/dev/ttyUSB0",
            "tflmi.arena_size": 2**17,  # esp32c3 ram ~300kB
            "tvmaot.arena_size": 2**17,
        }
    ],
)
def test_target_espidf(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    _, artifacts = _test_run_platform(
        "espidf", backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )

    assert len(lookup_artifacts(artifacts, name="generic_mlonmcu")) == 1
    assert len(lookup_artifacts(artifacts, name="espidf_out.log")) == 1


@pytest.mark.slow
@pytest.mark.hardware
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("backend_name", ["tvmllvm"])
@pytest.mark.parametrize("target_name", ["tvm_cpu"])
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize(
    "config",
    [{}],
)
def test_target_tvm(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    _, artifacts = _test_run_platform(
        "tvm", backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )

    assert len(lookup_artifacts(artifacts, name="tvmc_run_out.log")) == 1



import pytest
from mlonmcu.environment.config import PathConfig

from mlonmcu.session.run import RunStage
from mlonmcu.testing.helpers import DEFAULT_TFLITE_MODELS, DEFAULT_MLIF_TARGETS, _test_run_platform, _test_backend

# Run

# Frontend

# Backend + ?


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize("backend_name", ["tflmi"])
@pytest.mark.parametrize("target_name", DEFAULT_MLIF_TARGETS)
@pytest.mark.parametrize(
    "feature_names", [["muriscvnn"], ["muriscvnn", "debug"]]
)  # currently validate does not imply debug
@pytest.mark.parametrize("config", [{}])
def test_feature_muriscvnn(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    df, artifacts = _test_run_platform(
        None, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )
    assert "muriscvnn" in df["Features"][0]
    # TODO: find out if kernels are actually linked?
    # TODO: 2 runs to compare ROM/RAM/Cycles?


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize(
    "backend_name", ["tflmi"]  # -> add tvm if we have a test model for this
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("target_name", DEFAULT_MLIF_TARGETS)  # TODO: more targets (without vext)
@pytest.mark.parametrize(
    "feature_names", [["cmsisnn"], ["cmsisnn", "debug"]]
)  # currently validate does not imply debug
@pytest.mark.parametrize("config", [{}])
def test_feature_cmsisnn(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    df, artifacts = _test_run_platform(
        None, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )
    assert "cmsisnn" in df["Features"][0]
    # TODO: find out if kernels are actually linked?
    # TODO: 2 runs to compare ROM/RAM/Cycles?


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize(
    "backend_name", ["tvmaot"]  # -> add tvm if we have a test model for this
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("target_name", DEFAULT_MLIF_TARGETS)
@pytest.mark.parametrize(
    "feature_names", [["muriscvnnbyoc"], ["muriscvnnbyoc", "debug"]]
)  # currently validate does not imply debug
@pytest.mark.parametrize("config", [{}])
def test_feature_muriscvnnbyoc(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    df, artifacts = _test_run_platform(
        None, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )
    assert "muriscvnnbyoc" in df["Features"][0]
    # TODO: find out if kernels are actually linked?
    # TODO: 2 runs to compare ROM/RAM/Cycles?


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize(
    "backend_name", ["tvmaot"]  # -> add tvm if we have a test model for this
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("target_name", DEFAULT_MLIF_TARGETS)  # TODO: more targets (without vext)
@pytest.mark.parametrize(
    "feature_names", [["cmsisnnbyoc"], ["cmsisnnbyoc", "debug"]]
)  # currently validate does not imply debug
@pytest.mark.parametrize("config", [{}])
def test_feature_cmsisnnbyoc(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    df, artifacts = _test_run_platform(
        None, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )
    assert "cmsisnnbyoc" in df["Features"][0]
    # TODO: find out if kernels are actually linked?
    # TODO: 2 runs to compare ROM/RAM/Cycles?


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize(
    "backend_name", ["tvmaot"]  # -> add tvm if we have a test model for this
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("feature_names", [["unpacked_api"]])
@pytest.mark.parametrize("config", [{}])
def test_feature_unpacked_api(user_context, model_name, backend_name, models_dir, feature_names, config):
    df, artifacts = _test_backend(backend_name, user_context, model_name, models_dir, feature_names, config)
    assert "unpacked_api" in df["Features"][0]
    # TODO: check generated code -> do not run at all (would need to check for metrics changes)


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize(
    "backend_name", ["tvmaot"]  # -> add tvm if we have a test model for this
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("feature_names", [["usmp"]])
@pytest.mark.parametrize(
    "config",
    [{"usmp.algorithm": "greedy_by_size"}, {"usmp.algorithm": "greedy_by_conflicts"}, {"usmp.algorithm": "hill_climb"}],
)
def test_feature_usmp(user_context, model_name, backend_name, models_dir, feature_names, config):
    df, artifacts = _test_backend(backend_name, user_context, model_name, models_dir, feature_names, config)
    assert "usmp" in df["Features"][0]
    # TODO: run twice and compare generted code or look for specific code


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize(
    "backend_name", ["tvmaot"]  # other tvm backends?
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("feature_names", [["disable_legalize"]])
@pytest.mark.parametrize("config", [{}])
def test_feature_disable_legalize(user_context, model_name, backend_name, models_dir, feature_names, config):
    df, artifacts = _test_backend(backend_name, user_context, model_name, models_dir, feature_names, config)
    assert "disable_legalize" in df["Features"][0]
    # TODO: run twice and compare codegen results


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", ["sine_model"])  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize("frontend_name", ["tflite"])  # Validate is frontend feature as well
@pytest.mark.parametrize(
    "backend_name", ["tvmllvm"]  # other tvm backends?
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("target_name", ["tvm_cpu"])
@pytest.mark.parametrize("feature_names", [["autotvm"]])
@pytest.mark.parametrize("config", [{}])
def test_feature_autotvm(
    user_context, frontend_name, model_name, backend_name, target_name, models_dir, feature_names, config
):
    platform_name = "tvm"
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    if not user_context.environment.has_backend(backend_name):
        pytest.skip(f"Backend '{backend_name}' is not enabled.")
    if not user_context.environment.has_platform(platform_name):
        pytest.skip(f"Platform '{platform_name}' is not enabled.")  # TODO: not enabled -> not installed
    if not user_context.environment.has_target(target_name):
        pytest.skip(f"Target '{target_name}' is not enabled.")  # TODO: remove check?
    for feature in feature_names:
        if not user_context.environment.has_feature(feature):
            pytest.skip(f"Feature '{feature}' is not enabled.")
    user_context.environment.paths["models"] = [PathConfig(models_dir)]
    with user_context.create_session() as session:
        user_config = user_context.environment.vars.copy()
        user_config.update(config)
        run = session.create_run(config=user_config)
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        run.add_platform_by_name(platform_name, context=user_context)
        run.add_target_by_name(target_name, context=user_context)
        run.add_backend_by_name(backend_name, context=user_context)
        success = session.process_runs(until=RunStage.TUNE, context=user_context)
    report = session.get_reports()
    df = report.df
    assert success
    assert len(df) == 1
    assert "autotune" in df["Features"][0]


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", ["sine_model"])  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize("frontend_name", ["tflite"])  # Validate is frontend feature as well
@pytest.mark.parametrize(
    "backend_name", ["tvmllvm"]  # other tvm backends?
)  # TODO: Single backend would be fine, but it has to be enabled
@pytest.mark.parametrize("target_name", ["tvm_cpu"])
@pytest.mark.parametrize("feature_names", [["autotvm", "autotuned"]])  # TODO: provide tuning records instead
@pytest.mark.parametrize("config", [{"tvmaot.print_outputs": True}])
def test_feature_autotuned(
    user_context, frontend_name, model_name, backend_name, target_name, models_dir, feature_names, config, tmp_path
):
    platform_name = "tvm"
    if not user_context.environment.has_frontend(frontend_name):
        pytest.skip(f"Frontend '{frontend_name}' is not enabled.")
    if not user_context.environment.has_backend(backend_name):
        pytest.skip(f"Backend '{backend_name}' is not enabled.")
    if not user_context.environment.has_platform(platform_name):
        pytest.skip(f"Platform '{platform_name}' is not enabled.")  # TODO: not enabled -> not installed
    if not user_context.environment.has_target(target_name):
        pytest.skip(f"Target '{target_name}' is not enabled.")  # TODO: remove check?
    for feature in feature_names:
        if not user_context.environment.has_feature(feature):
            pytest.skip(f"Feature '{feature}' is not enabled.")
    user_context.environment.paths["models"] = [PathConfig(models_dir)]
    with user_context.create_session() as session:
        results_file = tmp_path / "tuning.log"
        results_file.touch()
        config.update({"autotuned.results_file": results_file})
        user_config = user_context.environment.vars.copy()
        user_config.update(config)
        run = session.create_run(config=user_config)
        run.add_features_by_name(feature_names, context=user_context)
        run.add_frontend_by_name(frontend_name, context=user_context)
        run.add_platform_by_name(platform_name, context=user_context)
        run.add_target_by_name(target_name, context=user_context)
        run.add_model_by_name(model_name, context=user_context)
        run.add_backend_by_name(backend_name, context=user_context)
        success = session.process_runs(until=RunStage.BUILD, context=user_context)
    report = session.get_reports()
    df = report.df
    assert success
    assert len(df) == 1
    assert "autotuned" in df["Features"][0]


# Platform
@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize(
    "model_name", DEFAULT_TFLITE_MODELS
)  # TODO: add test model for this, also test with wrong data
@pytest.mark.parametrize("backend_name", ["tflmi", "tvmaot"])
@pytest.mark.parametrize("target_name", ["host_x86"])
@pytest.mark.parametrize("feature_names", [["debug"]])  # currently validate does not imply debug
@pytest.mark.parametrize("config", [{}])
def test_feature_debug(user_context, model_name, backend_name, target_name, models_dir, feature_names, config):
    df, artifacts = _test_run_platform(
        None, backend_name, target_name, user_context, model_name, models_dir, feature_names, config
    )
    assert "debug" in df["Features"][0]
    # TODO: stdout with test model
    # TODO: 2 runs to compare ROM/RAM/Cycles?


# Target
# TODO

# TODO:
# cmsisnn -> currently broken
# gdbserver -> hard to test



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_muriscvnn(context: MlonMcuContext, params=None):
    if not context.environment.supports_feature("muriscvnn"):
        return False
    user_vars = context.environment.vars
    if "muriscvnn.src_dir" not in user_vars:
        assert "muriscvnn" in context.environment.repos, "Undefined repository: 'muriscvnn'"
    if params:
        toolchain = params.get("toolchain", "gcc")
        if not context.environment.has_toolchain(toolchain):
            return False
        target_arch = params.get("target_arch", "riscv")
        if target_arch == "riscv":
            if params.get("vext", False):
                if not context.environment.supports_feature("vext"):
                    return False
            if params.get("pext", False):
                if toolchain == "llvm":
                    # Unsupported
                    return False
                if not context.environment.supports_feature("pext"):
                    return False
            if params.get("vext", False) and params.get("pext", False):
                # Either pext or vext!
                return False
        elif target_arch == "x86":
            if toolchain != "gcc":
                return False
            if params.get("vext", False) or params.get("pext", False):
                return False
            # TODO: validate chosen toolchain?
    return True


def _validate_muriscvnn_build(context: MlonMcuContext, params=None):
    if not _validate_muriscvnn(context, params=params):
        return False
    user_vars = context.environment.vars
    skip_build = user_vars.get("muriscvnn.skip_build", True)
    # TODO: str2bool
    if skip_build:
        return False
    return True


@Tasks.provides(["muriscvnn.src_dir", "muriscvnn.inc_dir"])
@Tasks.validate(_validate_muriscvnn)
@Tasks.register(category=TaskType.OPT)
def clone_muriscvnn(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the muRISCV-NN project."""
    muriscvnnName = utils.makeDirName("muriscvnn")
    user_vars = context.environment.vars
    # if "muriscvnn.lib" in user_vars:  # TODO: also check command line flags?
    #     return False
    if "muriscvnn.src_dir" in user_vars:
        muriscvnnSrcDir = user_vars["muriscvnn.src_dir"]
        muriscvnnIncludeDir = Path(muriscvnnSrcDir) / "Include"
        rebuild = False
    else:
        muriscvnnSrcDir = context.environment.paths["deps"].path / "src" / muriscvnnName
        muriscvnnIncludeDir = muriscvnnSrcDir / "Include"
    if rebuild or not utils.is_populated(muriscvnnSrcDir):
        muriscvnnRepo = context.environment.repos["muriscvnn"]
        utils.clone_wrapper(muriscvnnRepo, muriscvnnSrcDir, refresh=rebuild)
    context.cache["muriscvnn.src_dir"] = muriscvnnSrcDir
    context.cache["muriscvnn.inc_dir"] = muriscvnnIncludeDir


@Tasks.needs(["muriscvnn.src_dir", "riscv_gcc.install_dir", "riscv_gcc.name"])
# @Tasks.optional(["riscv_gcc.install_dir", "riscv_gcc.name", "arm_gcc.install_dir"])
@Tasks.provides(["muriscvnn.build_dir", "muriscvnn.lib"])
# @Tasks.param("dbg", [False, True])
@Tasks.param("dbg", [False])  # disable due to bug with vext gcc
@Tasks.param("vext", [False])
@Tasks.param("pext", [False])
@Tasks.param("toolchain", ["gcc"])
@Tasks.param("target_arch", ["x86", "riscv"])
@Tasks.validate(_validate_muriscvnn_build)
@Tasks.register(category=TaskType.OPT)
def build_muriscvnn(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build muRISCV-NN."""
    if not params:
        params = {}
    flags = utils.makeFlags(
        (params["dbg"], "dbg"),
        (params["vext"], "vext"),
        (params["pext"], "pext"),
        (True, params["toolchain"]),
        (True, params["target_arch"]),
    )
    flags_ = utils.makeFlags((params["vext"], "vext"), (params["pext"], "pext"))
    muriscvnnName = utils.makeDirName("muriscvnn", flags=flags)
    muriscvnnSrcDir = context.cache["muriscvnn.src_dir"]
    muriscvnnBuildDir = context.environment.paths["deps"].path / "build" / muriscvnnName
    muriscvnnInstallDir = context.environment.paths["deps"].path / "install" / muriscvnnName
    muriscvnnLib = muriscvnnInstallDir / "libmuriscvnn.a"
    user_vars = context.environment.vars
    if "muriscvnn.lib" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not (utils.is_populated(muriscvnnBuildDir) and muriscvnnLib.is_file()):
        utils.mkdirs(muriscvnnBuildDir)
        muriscvnnArgs = []
        target_arch = params.get("target_arch", "riscv")
        if target_arch == "riscv":
            gccName = context.cache["riscv_gcc.name", flags_]
            toolchain = params.get("toolchain", "gcc")
            assert (
                gccName == "riscv32-unknown-elf" or toolchain != "llvm"
            ), "muRISCV-NN requires a non-multilib toolchain!"
            if "riscv_gcc.install_dir" in user_vars:
                riscv_gcc = user_vars["riscv_gcc.install_dir"]
            else:
                riscv_gcc = context.cache["riscv_gcc.install_dir", flags_]
            muriscvnnArgs.append("-DRISCV_GCC_PREFIX=" + str(riscv_gcc))
            muriscvnnArgs.append("-DTOOLCHAIN=" + params["toolchain"].upper())
            vext = params.get("vext", False)
            pext = params.get("pext", False)
            muriscvnnArgs.append("-DUSE_VEXT=" + ("ON" if vext else "OFF"))
            muriscvnnArgs.append("-DUSE_PEXT=" + ("ON" if pext else "OFF"))
            muriscvnnArgs.append(f"-DRISCV_GCC_BASENAME={gccName}")
            arch = "rv32imafdc"
            if vext:
                arch += "v"
            if pext:
                arch += "p"
            muriscvnnArgs.append(f"-DRISCV_ARCH={arch}")
        elif target_arch == "x86":
            toolchain = params.get("toolchain", "gcc")
            muriscvnnArgs.append("-DTOOLCHAIN=x86")
            muriscvnnArgs.append("-DUSE_VEXT=OFF")
            muriscvnnArgs.append("-DUSE_PEXT=OFF")
        else:
            raise RuntimeError(f"Unsupported target_arch for muriscvnn: {target_arch}")
        utils.cmake(
            muriscvnnSrcDir,
            *muriscvnnArgs,
            cwd=muriscvnnBuildDir,
            debug=params["dbg"],
            live=verbose,
        )
        utils.make(cwd=muriscvnnBuildDir, threads=threads, live=verbose)
        utils.mkdirs(muriscvnnInstallDir)
        utils.move(muriscvnnBuildDir / "Source" / "libmuriscvnn.a", muriscvnnLib)
    context.cache["muriscvnn.build_dir", flags] = muriscvnnBuildDir
    context.cache["muriscvnn.lib", flags] = muriscvnnLib



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os


def prepare_python_environment(
    pythonpath, tvm_build_dir, tvm_configs_dir, tophub_url=None, num_threads=None, debug_cfg=None
):
    env = os.environ.copy()
    if debug_cfg:
        env["TVM_LOG_DEBUG"] = debug_cfg
    if pythonpath:
        env["PYTHONPATH"] = str(pythonpath)
    if tvm_build_dir:
        env["TVM_LIBRARY_PATH"] = str(tvm_build_dir)
    if tvm_configs_dir:
        env["TVM_CONFIGS_JSON_DIR"] = str(tvm_configs_dir)
    if tophub_url:
        env["TOPHUB_LOCATION"] = tophub_url
    if num_threads:
        env["TVM_NUM_THREADS"] = str(num_threads)
    # Use all cores/threads for building models and let OS scheduler decide on the mapping
    env["TVM_BIND_THREADS"] = str(0)
    return env



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions for TVMFramework."""

import os
import pkg_resources
from pathlib import Path

from mlonmcu.flow.framework import Framework

# from mlonmcu.flow.tvm import TVMBackend


def get_crt_config_dir():
    files = pkg_resources.resource_listdir(
        "mlonmcu", os.path.join("..", "resources", "frameworks", "tvm", "crt_config")
    )
    if "crt_config.h" not in files:
        return None
    fname = pkg_resources.resource_filename(
        "mlonmcu", os.path.join("..", "resources", "frameworks", "tvm", "crt_config")
    )
    return fname


class TVMFramework(Framework):
    """TVM Framework specialization."""

    name = "tvm"

    FEATURES = {"cmsisnnbyoc", "muriscvnnbyoc"}

    DEFAULTS = {
        "extra_incs": [],
        "extra_libs": [],
        "crt_config_dir": get_crt_config_dir(),
    }

    REQUIRED = {"tvm.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(features=features, config=config)
        # self.backends = TVMBackend.registry

    @property
    def tvm_src(self):
        return Path(self.config["tvm.src_dir"])

    @property
    def extra_incs(self):
        return self.config["extra_incs"]

    @property
    def extra_libs(self):
        return self.config["extra_libs"]

    @property
    def crt_config_dir(self):
        return self.config["crt_config_dir"]

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        if self.extra_incs or self.extra_libs:
            assert platform == "mlif", "Extra incs or libs are only supported by 'mlif' platform"
            if self.extra_incs:
                if isinstance(self.extra_incs, list):
                    temp = r"\;".join(self.extra_incs)
                else:
                    temp = self.extra_incs
                ret["TVM_EXTRA_INCS"] = temp
            if self.extra_libs:
                if isinstance(self.extra_libs, list):
                    temp = r"\;".join(self.extra_libs)
                else:
                    temp = self.extra_libs
                ret["TVM_EXTRA_LIBS"] = temp
        if self.crt_config_dir:
            ret["TVM_CRT_CONFIG_DIR"] = self.crt_config_dir
        ret["TVM_DIR"] = str(self.tvm_src)
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for exporting session and runs."""
from mlonmcu.context.context import MlonMcuContext

from mlonmcu.cli.common import (
    add_common_options,
    add_context_options,
)


def add_export_options(parser):
    export_parser = parser.add_argument_group("export options")
    export_parser.add_argument(
        "destination",
        nargs="?",
        default="exported.zip",
        help="Path to the output directory or archive (default: %(default)s)",
    )
    export_parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Overwrite files if the destination already exists (DANGEROUS)",
    )
    export_parser.add_argument(
        "-l",
        "--list",
        default=False,
        action="store_true",
        help="Print a summary of all available sessions and runs",
    )
    export_parser.add_argument(
        "-s",
        "--session",
        metavar="SESSION",
        type=int,
        nargs="?",
        default=None,
        const=-1,
        action="append",
        help="Which session(s) should be exported (default: latest session id)",
    )
    export_parser.add_argument(
        "-r",
        "--run",
        metavar="RUN",
        type=int,
        nargs="?",
        default=None,
        const=-1,
        action="append",
        help="Which run(s) should be exported (default: all runs of the selected session/latest run id)",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the cleanup subcommand."""
    parser = subparsers.add_parser("export", description="Export session/run artifacts to a directory/archive.")
    parser.set_defaults(func=handle)
    add_common_options(parser)
    add_context_options(parser)
    add_export_options(parser)
    return parser


def handle(args):
    with MlonMcuContext(path=args.home, deps_lock="read") as context:
        if args.list:
            context.print_summary(sessions=True, runs=True, labels=True)
            return 0
        dest = args.destination
        interactive = not args.force
        sids = args.session
        rids = args.run
        context.export(dest, session_ids=sids, run_ids=rids, interactive=interactive)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import mock
import pytest
import yaml
import re
from mlonmcu.environment.environment import PathConfig
from mlonmcu.models.lookup import print_summary

# def test_models_get_model_directories():
#     pass
#
# def test_models_find_metadata():
#     pass
#
# def test_models_list_model_subdir():  # TODO
#     pass
#
# def test_models_list_models():
#     pass
#
# def test_models_list_modelgroups():
#     pass
#
# def lookup_models_and_groups():
#     pass

# def test_models_print_paths():
#     pass
#
# def test_models_print_models():
#     pass
#
# def test_models_print_groups():
#     pass


def _create_fake_metadata(path, base="metadata", ext="yaml"):
    metadata_file = path / f"{base}.{ext}"
    dummy_data = {}
    with open(metadata_file.resolve(), "w") as handle:
        yaml.dump(dummy_data, handle)


def _create_fake_models(path, names, ext="tflite", with_subdir=True, with_metadata=False):
    for name in names:
        if with_subdir:
            path.mkdir(exist_ok=True)
            model_dir = path / name
        else:
            model_dir = path
        model_dir.mkdir()
        model_file = model_dir / f"{name}.{ext}"
        model_file.touch()
        if with_metadata:
            if with_subdir:
                _create_fake_metadata(model_dir)
            else:
                _create_fake_metadata(model_dir, base=name)


def _create_fake_modelgroups(path, data={}, base="groups", ext="yaml"):
    path.mkdir(exist_ok=True)
    groups_file = path / f"{base}.{ext}"

    with open(groups_file.resolve(), "w") as handle:
        yaml.dump(data, handle)


def _check_summary_output(
    out,
    expected_paths=[],
    expected_models=[],
    expected_groups={},
    expected_skipped=0,
    expected_metadata_count=0,
    expected_duplicates=0,
    expected_group_duplicates=0,
    detailed=False,
):
    expected_paths_str = ["    " + str(path) for path in expected_paths]
    expected_models_str = ["    " + name for name in expected_models]
    expected_groups_str = [
        "    " + name + " [{} models]".format(len(models)) for name, models in expected_groups.items()
    ]
    skipped_count = 0
    duplicates_count = 0
    metadata_count = 0
    for line in out.split("\n"):
        if "(skipped)" in line:
            skipped_count = skipped_count + 1
            line = line.replace(" (skipped)", "")
        match = re.compile(r".*\((\d*) duplicate[s]*\).*").search(line)
        if match:
            count = int(match.group(1))
            duplicates_count = duplicates_count + count
            line = line[: line.find(" (")]

        if detailed:
            if line.startswith("        "):
                if "Metadata: available" in line:
                    metadata_count = metadata_count + 1
                continue
        assert (
            line
            in ["Models Summary", "Paths:", "Models:", "Groups:", ""]
            + expected_paths_str
            + expected_models_str
            + expected_groups_str
        )
    assert skipped_count == expected_skipped
    assert duplicates_count == expected_duplicates + expected_group_duplicates
    if detailed:
        assert metadata_count == expected_metadata_count


@pytest.mark.parametrize("detailed", [False, True])
def test_models_print_summary(detailed, capsys, fake_context, fake_environment_directory, fake_config_home):
    # list empty
    with mock.patch.dict(fake_context.environment.paths, {"models": []}):
        print_summary(fake_context, detailed=detailed)
        out, err = capsys.readouterr()
        for line in out.split("\n"):
            assert line in ["Models Summary", "Paths:", "Models:", "Groups:", ""]

    env_models_dir = fake_environment_directory / "models"
    env_models_dir.mkdir()

    # single path, no models
    with mock.patch.dict(fake_context.environment.paths, {"models": [PathConfig(env_models_dir)]}):
        print_summary(fake_context, detailed=detailed)
        out, err = capsys.readouterr()
        _check_summary_output(
            out,
            detailed=detailed,
            expected_paths=[env_models_dir],
        )

    _create_fake_models(env_models_dir, ["model0"], with_metadata=True)
    _create_fake_models(env_models_dir, ["model1", "model2"])

    # single path, 3 models
    with mock.patch.dict(fake_context.environment.paths, {"models": [PathConfig(env_models_dir)]}):
        print_summary(fake_context, detailed=detailed)
        out, err = capsys.readouterr()
        _check_summary_output(
            out,
            detailed=detailed,
            expected_paths=[env_models_dir],
            expected_models=["model0", "model1", "model2"],
            expected_metadata_count=1,
        )

    user_models_dir = fake_config_home / "models"

    # two paths (1 skipped), 3 models
    with mock.patch.dict(
        fake_context.environment.paths, {"models": [PathConfig(env_models_dir), PathConfig(user_models_dir)]}
    ):
        print_summary(fake_context, detailed=detailed)
        out, err = capsys.readouterr()
        _check_summary_output(
            out,
            detailed=detailed,
            expected_paths=[env_models_dir, user_models_dir],
            expected_models=["model0", "model1", "model2"],
            expected_skipped=1,
            expected_metadata_count=1,
        )

    user_models_dir.mkdir()
    _create_fake_models(fake_config_home / "models", ["model2", "model3"])

    # two paths, 5 models, 1 duplicate
    with mock.patch.dict(
        fake_context.environment.paths,
        {"models": [PathConfig(fake_environment_directory / "models"), PathConfig(fake_config_home / "models")]},
    ):
        print_summary(fake_context, detailed=detailed)
        out, err = capsys.readouterr()
        _check_summary_output(
            out,
            detailed=detailed,
            expected_paths=[env_models_dir, user_models_dir],
            expected_models=["model0", "model1", "model2", "model3"],
            expected_metadata_count=1,
            expected_duplicates=1,
        )

    _create_fake_modelgroups(env_models_dir, {"mygroup": ["model0", "model2"]})

    # two paths, 5 models, 1 duplicate, 1 group
    with mock.patch.dict(
        fake_context.environment.paths,
        {"models": [PathConfig(fake_environment_directory / "models"), PathConfig(fake_config_home / "models")]},
    ):
        print_summary(fake_context, detailed=detailed)
        out, err = capsys.readouterr()
        print("OUT", out)
        _check_summary_output(
            out,
            detailed=detailed,
            expected_paths=[env_models_dir, user_models_dir],
            expected_models=["model0", "model1", "model2", "model3"],
            expected_groups={"mygroup": ["model0", "model2"]},
            expected_metadata_count=1,
            expected_duplicates=1,
        )
    # TODO: group name conflicts with modelname
    # TODO: duplicate groups



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
import pytest
from mlonmcu.context.read_write_filelock import ReadFileLock, WriteFileLock, RWLockTimeout


@pytest.mark.parametrize("test_input_raise_exception", [False, True])
def test_read_after_read(monkeypatch, fake_environment_directory: Path, test_input_raise_exception):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    readlock1 = ReadFileLock(filepath)
    print("read1: " + readlock1.id)
    readlock2 = ReadFileLock(filepath)
    print("read2: " + readlock2.id)

    assert readlock1.acquire(raise_exception=test_input_raise_exception)
    assert readlock2.acquire(raise_exception=test_input_raise_exception)

    readlock1.release()
    readlock2.release()


def test_read_after_write_exception_disabled(monkeypatch, fake_environment_directory: Path):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    writelock1 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)
    readlock1 = ReadFileLock(filepath)
    print("read1: " + readlock1.id)
    assert writelock1.acquire(raise_exception=False)
    assert not readlock1.acquire(raise_exception=False)

    writelock1.release()


def test_read_after_write_exception_enabled(monkeypatch, fake_environment_directory: Path):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    writelock1 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)
    readlock1 = ReadFileLock(filepath)
    print("read1: " + readlock1.id)

    assert writelock1.acquire()
    with pytest.raises(RWLockTimeout, match=r".*The lock.*could\ not\ be\ acquired\..*"):
        readlock1.acquire()

    writelock1.release()


def test_write_after_read_exception_disabled(monkeypatch, fake_environment_directory: Path):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    readlock1 = ReadFileLock(filepath)
    print("read1: " + readlock1.id)
    writelock1 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)
    assert readlock1.acquire(raise_exception=False)
    assert not writelock1.acquire(raise_exception=False)

    readlock1.release()


def test_write_after_read_exception_enabled(monkeypatch, fake_environment_directory: Path):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    readlock1 = ReadFileLock(filepath)
    print("read1: " + readlock1.id)
    writelock1 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)

    assert readlock1.acquire()
    with pytest.raises(RWLockTimeout, match=r".*The lock.*could\ not\ be\ acquired\..*"):
        writelock1.acquire()

    readlock1.release()


def test_write_after_write_exception_disabled(monkeypatch, fake_environment_directory: Path):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    writelock1 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)
    writelock2 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)
    assert writelock1.acquire(raise_exception=False)
    assert not writelock2.acquire(raise_exception=False)

    writelock1.release()


def test_write_after_write_exception_enabled(monkeypatch, fake_environment_directory: Path):
    monkeypatch.chdir(fake_environment_directory)
    filepath = fake_environment_directory / ".lock"

    writelock1 = WriteFileLock(filepath)
    print("write1: " + writelock1.id)
    writelock2 = WriteFileLock(filepath)
    print("write1: " + writelock2.id)

    assert writelock1.acquire()
    with pytest.raises(RWLockTimeout, match=r".*The lock.*could\ not\ be\ acquired\..*"):
        writelock2.acquire()

    writelock1.release()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Spike Target definitions"""


from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from .riscv import RISCVTarget
from .util import update_extensions

logger = get_logger()


class RVBTarget(RISCVTarget):
    """TODO"""

    FEATURES = RISCVTarget.FEATURES | {"bext"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "enable_bext": False,
        "bext_spec": 0.92,
        "bext_zba": False,
        "bext_zbb": False,
        "bext_zbc": False,
        "bext_zbs": False,
    }
    REQUIRED = RISCVTarget.REQUIRED

    def __init__(
        self,
        name,
        features=None,
        config=None,
    ):
        super().__init__(name, features=features, config=config)

    @property
    def enable_bext(self):
        value = self.config["enable_bext"]
        return str2bool(value)

    @property
    def bext_spec(self):
        return float(self.config["bext_spec"])

    @property
    def bext_zba(self):
        value = self.config["bext_zba"]
        return str2bool(value)

    @property
    def bext_zbb(self):
        value = self.config["bext_zbb"]
        return str2bool(value)

    @property
    def bext_zbc(self):
        value = self.config["bext_zbc"]
        return str2bool(value)

    @property
    def bext_zbs(self):
        value = self.config["bext_zbs"]
        return str2bool(value)

    @property
    def extensions(self):
        exts = super().extensions
        return update_extensions(
            exts,
            bext=self.enable_pext,
            bext_spec=self.pext_spec,
            bext_zba=self.bext_zba,
            bext_zbb=self.bext_zbb,
            bext_zbc=self.bext_zbc,
            bext_zbs=self.bext_zbs,
        )

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        if self.enable_bext:
            major, minor = str(self.bext_spec).split(".")[:2]
            ret["RISCV_RVB_MAJOR"] = major
            ret["RISCV_RVB_MINOR"] = minor
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# init
import os
import mock
import pytest
import configparser
from mlonmcu.environment.init import (
    create_environment_directories,
    create_venv_directory,
    ask_user,
    initialize_environment,
)
from mlonmcu.environment.list import get_alternative_name, register_environment
from mlonmcu.environment.templates import fill_template, write_environment_yaml_from_template, get_template_text
from mlonmcu.environment.config import get_environments_file, init_config_dir


def test_environment_create_environment_directories(fake_environment_directory):
    directories = ["foo", "bar", "test"]
    create_environment_directories(fake_environment_directory, directories)
    assert all([(fake_environment_directory / d).is_dir() for d in directories])


# def test_environment_clone_models_repo(fake_environment_directory):
#     subdir = fake_environment_directory / "models"
#     clone_models_repo(subdir)
#     assert (subdir / ".git").is_dir()


@pytest.mark.parametrize("interactive", [False, True])
@pytest.mark.parametrize("default", [False, True])
def test_environment_ask_user(interactive, default):
    msg = "Answer?"
    # empty or implicit
    with mock.patch("builtins.input", return_value=""):
        assert ask_user(msg, default=default, interactive=interactive) == default
    with mock.patch("builtins.input", return_value="foo"):
        assert ask_user(msg, default=default, interactive=interactive) == default
    # yes (explicit)
    with mock.patch("builtins.input", return_value="y"):
        if interactive:
            assert ask_user(msg, default=default, interactive=interactive)
        else:
            assert ask_user(msg, default=default, interactive=interactive) == default
    with mock.patch("builtins.input", return_value="Y"):
        if interactive:
            assert ask_user(msg, default=default, interactive=interactive)
        else:
            assert ask_user(msg, default=default, interactive=interactive) == default
    # no  (explicit)
    with mock.patch("builtins.input", return_value="n"):
        if interactive:
            assert not ask_user(msg, default=default, interactive=interactive)
        else:
            assert ask_user(msg, default=default, interactive=interactive) == default
    with mock.patch("builtins.input", return_value="N"):
        if interactive:
            assert not ask_user(msg, default=default, interactive=interactive)
        else:
            assert ask_user(msg, default=default, interactive=interactive) == default


@pytest.mark.parametrize("hidden", [False, True])
def test_environment_create_venv_directory(hidden, fake_environment_directory):
    create_venv_directory(fake_environment_directory, hidden=hidden)
    assert (fake_environment_directory / (".venv" if hidden else "venv")).is_dir()


def _count_envs(directory):
    envs_file = directory / "environments.ini"
    parser = configparser.ConfigParser()
    parser.read(str(envs_file))
    return len(parser.sections())


def _has_duplicate_envs(directory):
    envs_file = directory / "environments.ini"
    parser = configparser.ConfigParser()
    parser.read(str(envs_file))
    secs = parser.sections()
    unique = list(dict.fromkeys(secs))  # Without duplicates
    return len(unique) != len(secs)


def test_environment_get_environment_names():
    pass  # TODO


def test_environment_get_alternative_name():
    assert get_alternative_name("foo", ["foo", "bar", "foo_bar"]) == "foo_0"
    assert get_alternative_name("foo", ["foo", "bar", "foo_bar", "foo_0"]) == "foo_1"
    assert get_alternative_name("foo_9", ["foo_9", "foo_10", "foo_9_0"]) == "foo_9_1"


def test_environment_register_environment(
    fake_config_home,
):
    pass


def test_environment_register_environment_invalid(
    fake_config_home,
):
    init_config_dir()
    register_environment("foo", "/a/b/c", False)

    # already exists
    with pytest.raises(RuntimeError):
        register_environment("foo", "/a/b/c", False)
    register_environment("foo", "/a/b/c", True)

    # rel path
    with pytest.raises(RuntimeError):
        register_environment("bar", "a", False)

    # missing envs file
    os.remove(get_environments_file())
    with pytest.raises(RuntimeError):
        register_environment("bar", "/a/b/c", True)


@pytest.mark.parametrize("interactive", [False])  # We do not want to mock the user input here
@pytest.mark.parametrize("create_venv", [False])  # Already covered above
@pytest.mark.parametrize("clone_models", [False])  # Already covered above
@pytest.mark.parametrize("register", [False, True])  #
@pytest.mark.parametrize("template", ["default"])
def test_environment_initialize_environment(
    interactive,
    create_venv,
    clone_models,
    register,
    template,
    fake_environment_directory,
    fake_working_directory,
    fake_config_home,
):
    with mock.patch("mlonmcu.environment.templates.get_template_text", return_value=str.encode("---")):
        before = _count_envs(fake_config_home)
        initialize_environment(
            fake_environment_directory,
            "myenv",
            interactive=interactive,
            create_venv=create_venv,
            clone_models=clone_models,
            register=register,
            template=template,
        )
        initialize_environment(
            fake_working_directory,
            "another",
            interactive=interactive,
            create_venv=create_venv,
            clone_models=clone_models,
            register=register,
            template=template,
        )
        assert _count_envs(fake_config_home) == ((before + 2) if register else before)


def test_environment_initialize_environment_duplicate(
    fake_environment_directory, fake_working_directory, fake_config_home
):
    with mock.patch("mlonmcu.environment.templates.get_template_text", return_value=str.encode("---")):
        before = _count_envs(fake_config_home)
        initialize_environment(
            fake_environment_directory,
            "",
            interactive=False,
            create_venv=False,
            clone_models=False,
            register=True,
            template="default",
        )
        initialize_environment(
            fake_working_directory,
            "",
            interactive=False,
            create_venv=False,
            clone_models=False,
            register=True,
            template="default",
        )
        assert _count_envs(fake_config_home) == (before + 2)
        assert not _has_duplicate_envs(fake_config_home)


# template


def test_environment_get_template_text():
    assert len(get_template_text("default")) > 0


def test_environment_fill_template(tmp_path_factory):
    # by file name
    fname = tmp_path_factory.mktemp("dir") / "file.yml.j2"
    with open(fname, "w") as f:
        f.write("{{ key }}: {{ value }}\n")
    assert fill_template(fname, data={"key": "foo", "value": "bar"}) == "foo: bar"

    # by template name
    with mock.patch(
        "mlonmcu.environment.templates.get_template_text", return_value=str.encode("{{ key }}: {{ value }}")
    ):
        assert fill_template("", data={"key": "foo", "value": "bar"}) == "foo: bar"
    with mock.patch("mlonmcu.environment.templates.get_template_text", return_value=None):
        assert fill_template("", data={"key": "foo", "value": "bar"}) is None

    # error handling
    with pytest.raises(UnicodeDecodeError):
        with mock.patch("mlonmcu.environment.templates.get_template_text", return_value=b"\xff"):
            assert fill_template("")


def test_environment_write_environment_yaml_from_template(fake_environment_directory, fake_config_home):
    with mock.patch(
        "mlonmcu.environment.templates.get_template_text",
        return_value=str.encode("home: {{ home_dir }}\nconfig: {{ config_dir }}"),
    ):
        env_yaml = str(fake_environment_directory / "environment.yml")
        write_environment_yaml_from_template(env_yaml, "", fake_environment_directory)
        assert os.path.isfile(env_yaml)
        with open(env_yaml) as yaml:
            lines = "".join(yaml.readlines())
            assert lines == f"home: {fake_environment_directory}\nconfig: {fake_config_home}"



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Spike Target definitions"""

import os
import re
import time
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.bench import add_bench_metrics
from .riscv_pext_target import RVPTarget
from .riscv_vext_target import RVVTarget
from .riscv_bext_target import RVBTarget
from .util import update_extensions, sort_extensions_canonical, join_extensions

logger = get_logger()


def filter_unsupported_extensions(exts):
    assert isinstance(exts, set)
    REPLACEMENTS = {
        r"zve\d\d[xfd]": "v",
        r"zvl\d+b": None,
        r"zpsfoperand": "p",
        r"zpn": "p",
        r"zbpo": "p",
        # r"p": ["p", "b"],
        # r"p": ["p", "zba", "zbb", "zbc", "zbs"],
    }
    ret = set()
    for ext in exts:
        ignore = False
        for key, value in REPLACEMENTS.items():
            m = re.compile(key).match(ext)
            if m:
                if value:
                    if isinstance(value, list):
                        assert len(value) > 0
                        ret |= set(value)
                    else:
                        ret.add(value)
                ignore = True
        if not ignore:
            ret.add(ext)
    ret = set(ret)

    return ret


class SpikeTarget(RVPTarget, RVVTarget, RVBTarget):
    """Target using the riscv-isa-sim (Spike) RISC-V simulator."""

    FEATURES = RVPTarget.FEATURES | RVVTarget.FEATURES | RVBTarget.FEATURES | {"cachesim", "log_instrs"}

    DEFAULTS = {
        **RVPTarget.DEFAULTS,
        **RVVTarget.DEFAULTS,
        **RVBTarget.DEFAULTS,
        "spikepk_extra_args": [],
    }
    REQUIRED = RVPTarget.REQUIRED | RVVTarget.REQUIRED | RVBTarget.REQUIRED | {"spike.exe", "spike.pk"}

    def __init__(self, name="spike", features=None, config=None):
        super().__init__(name, features=features, config=config)

    @property
    def spike_exe(self):
        return Path(self.config["spike.exe"])

    @property
    def spike_pk(self):
        return Path(self.config["spike.pk"])

    @property
    def spikepk_extra_args(self):
        return self.config["spikepk_extra_args"]

    @property
    def extensions(self):
        # exts = RVPTarget.extensions(self) + RVVTarget.extensions(self)
        exts = super().extensions
        return update_extensions(
            exts,
        )

    @property
    def isa(self):
        exts = filter_unsupported_extensions(self.extensions)
        exts_str = join_extensions(sort_extensions_canonical(exts, lower=True))
        return f"rv{self.xlen}{exts_str}"

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        spike_args = []
        spikepk_args = []

        spike_args.append(f"--isa={self.isa}")

        if len(self.extra_args) > 0:
            if isinstance(self.extra_args, str):
                extra_args = self.extra_args.split(" ")
            else:
                extra_args = self.extra_args
            spike_args.extend(extra_args)

        spikepk_args.append("-s")

        if len(self.spikepk_extra_args) > 0:
            if isinstance(self.spikepk_extra_args, str):
                extra_args = self.spikepk_extra_args.split(" ")
            else:
                extra_args = self.spikepk_extra_args
            spikepk_args.extend(extra_args)  # I rename args to extra_args because otherwise it overwrites *args

        if self.enable_vext:
            assert self.vlen < 8192, "Spike does not support VLEN >= 8192"
            spike_args.append(f"--varch=vlen:{self.vlen},elen:{self.elen}")
        else:
            # assert self.vlen == 0
            pass

        if self.timeout_sec > 0:
            raise NotImplementedError

        ret = execute(
            self.spike_exe.resolve(),
            *spike_args,
            self.spike_pk.resolve(),
            *spikepk_args,
            program,
            *args,
            cwd=cwd,
            **kwargs,
        )
        return ret

    def parse_stdout(self, out, metrics, exit_code=0):
        add_bench_metrics(out, metrics, exit_code != 0, target_name=self.name)
        sim_insns = re.search(r"(\d*) cycles", out)
        if sim_insns:
            sim_insns = int(float(sim_insns.group(1)))
            metrics.add("Simulated Instructions", sim_insns, True)

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            # TODO: before or after?
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        start_time = time.time()
        if self.print_outputs:
            out = self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
        else:
            out = self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        # size instead of readelf?

        # TODO: get exit code
        exit_code = 0
        metrics = Metrics()
        self.parse_stdout(out, metrics, exit_code=exit_code)

        if metrics.has("Simulated Instructions"):
            sim_insns = metrics.get("Simulated Instructions")
            if diff > 0:
                metrics.add("MIPS", (sim_insns / diff) / 1e6, True)

        return metrics, out, []

    def get_platform_defs(self, platform):
        ret = {}
        ret.update(RVPTarget.get_platform_defs(self, platform))
        ret.update(RVVTarget.get_platform_defs(self, platform))
        ret.update(RVBTarget.get_platform_defs(self, platform))
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        if backend in SUPPORTED_TVM_BACKENDS:
            if optimized_layouts:
                if self.enable_pext or self.enable_vext:
                    ret.update(
                        {
                            # Warning: passing kernel layouts does not work with upstream TVM
                            # TODO: allow passing map?
                            "desired_layout": "NHWC:HWOI",
                        }
                    )
        return ret


if __name__ == "__main__":
    cli(target=SpikeTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys


def is_power_of_two(n):
    return (n & (n - 1) == 0) and n != 0


def ask_user(text, default: bool, yes_keys=["y", "j"], no_keys=["n"], interactive=True):
    assert len(yes_keys) > 0 and len(no_keys) > 0
    if not interactive:
        return default
    if default:
        suffix = " [{}/{}] ".format(yes_keys[0].upper(), no_keys[0].lower())
    else:
        suffix = " [{}/{}] ".format(yes_keys[0].lower(), no_keys[0].upper())
    message = text + suffix
    answer = input(message)
    if default:
        return answer.lower() not in no_keys and answer.upper() not in no_keys
    else:
        return not (answer.lower() not in yes_keys and answer.upper() not in yes_keys)


def get_base_prefix_compat():
    """Get base/real prefix, or sys.prefix if there is none."""
    return getattr(sys, "base_prefix", None) or getattr(sys, "real_prefix", None) or sys.prefix


def in_virtualenv():
    """Detects if the current python interpreter is from a virtual environment."""
    return get_base_prefix_compat() != sys.prefix


def filter_none(data):
    """Helper function which drop dict items with a None value."""
    assert isinstance(data, dict), "Dict only"
    out = {key: value for key, value in data.items() if value is not None}
    return out



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Flow module for frameworks and backend."""

from mlonmcu.flow.tflm.framework import TFLMFramework
from mlonmcu.flow.tvm.framework import TVMFramework

from mlonmcu.flow.tflm.backend.tflmc import TFLMCBackend
from mlonmcu.flow.tflm.backend.tflmi import TFLMIBackend

from mlonmcu.flow.tvm.backend.tvmaot import TVMAOTBackend
from mlonmcu.flow.tvm.backend.tvmaotplus import TVMAOTPlusBackend
from mlonmcu.flow.tvm.backend.tvmcg import TVMCGBackend
from mlonmcu.flow.tvm.backend.tvmllvm import TVMLLVMBackend
from mlonmcu.flow.tvm.backend.tvmrt import TVMRTBackend

SUPPORTED_FRAMEWORKS = {
    "tflm": TFLMFramework,
    "tvm": TVMFramework,
}

SUPPORTED_TFLITE_BACKENDS = {
    "tflmc": TFLMCBackend,
    "tflmi": TFLMIBackend,
}

SUPPORTED_TVM_BACKENDS = {
    "tvmaot": TVMAOTBackend,
    "tvmaotplus": TVMAOTPlusBackend,
    "tvmrt": TVMRTBackend,
    "tvmcg": TVMCGBackend,
    "tvmllvm": TVMLLVMBackend,
}

SUPPORTED_FRAMEWORK_BACKENDS = {
    "tflm": SUPPORTED_TFLITE_BACKENDS,
    "tvm": SUPPORTED_TVM_BACKENDS,
}

SUPPORTED_BACKENDS = {**SUPPORTED_TFLITE_BACKENDS, **SUPPORTED_TVM_BACKENDS}


def get_available_backend_names():
    """Return all available backend names."""
    return list(SUPPORTED_BACKENDS.keys())



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import stat
import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_corev_ovpsim(context: MlonMcuContext, params=None):
    return context.environment.has_target("corev_ovpsim")


@Tasks.provides(["corev_ovpsim.exe", "corev_ovpsim.install_dir"])
@Tasks.validate(_validate_corev_ovpsim)
@Tasks.register(category=TaskType.TARGET)
def install_corev_ovpsim(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install CORE-V OVPSim."""
    ovpName = utils.makeDirName("corev_ovpsim")
    ovpInstallDir = context.environment.paths["deps"].path / "install" / ovpName
    # TODO: windows?
    ovpExe = ovpInstallDir / "bin" / "Linux64" / "riscvOVPsimCOREV.exe"
    user_vars = context.environment.vars
    if "corev_ovpsim.exe" in user_vars:  # TODO: also check command line flags?
        return False
    else:
        if not ovpExe.is_file():

            def _helper(url):
                fullUrlSplit = url.split("/")
                tmpUrl = "/".join(fullUrlSplit[:-1])
                tmpFileName, tmpFileExtension = fullUrlSplit[-1].split(".", 1)
                return tmpUrl, tmpFileName, tmpFileExtension

            if "corev_ovpsim.dl_url" in user_vars:
                ovpUrl, ovpFileName, ovpFileExtension = _helper(user_vars["corev_ovpsim.dl_url"])
            else:
                ovpUrl = "https://github.com/openhwgroup/riscv-ovpsim-corev/archive/refs/heads/"
                # ovpVersion = user_vars.get("corev_ovpsim.version", "v20230425")
                ovpVersion = user_vars.get("corev_ovpsim.version", "v20230724")
                ovpFileName = str(ovpVersion)
                ovpFileExtension = "zip"
            ovpArchive = ovpFileName + "." + ovpFileExtension
            utils.download_and_extract(ovpUrl, ovpArchive, ovpInstallDir, progress=verbose)
            st = os.stat(ovpExe)
            # make executable
            os.chmod(ovpExe, st.st_mode | stat.S_IEXEC)
    context.cache["corev_ovpsim.install_dir"] = ovpInstallDir
    context.cache["corev_ovpsim.exe"] = ovpExe



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import yaml
import pathlib
import logging

from .config import (
    DefaultsConfig,
    PathConfig,
    RepoConfig,
    FrameworkConfig,
    FrameworkFeatureConfig,
    BackendConfig,
    BackendFeatureConfig,
    TargetConfig,
    TargetFeatureConfig,
    PlatformConfig,
    PlatformFeatureConfig,
    FrontendConfig,
    FrontendFeatureConfig,
)

# def load_environment_from_file(filename):
#     """Utility to initialize a mlonmcu environment from a YAML file."""
#     if isinstance(filename, str):
#         filename = pathlib.Path(filename)
#     with open(filename, encoding="utf-8") as yaml_file:
#         data = yaml.safe_load(yaml_file)
#         if data:
#             if "home" in data:
#                 print(data["home"], filename.parent)
#                 assert os.path.realpath(data["home"]) == os.path.realpath(filename.parent)
#             else:
#                 data["home"] = filename.parent
#             env = Environment(data)
#             return env
#         raise RuntimeError(f"Error opening environment file: {filename}")
#     return None


def load_environment_from_file(filename, base):
    """Utility to initialize a mlonmcu environment from a YAML file."""
    if isinstance(filename, str):
        filename = pathlib.Path(filename)
    with open(filename, encoding="utf-8") as yaml_file:
        loaded = yaml.safe_load(yaml_file)
        if not loaded:
            raise RuntimeError("Invalid YAML contents")
        if "home" in loaded:
            home = loaded["home"]
        else:
            home = None
        if "logging" in loaded:
            if "level" in loaded["logging"]:
                log_level = logging.getLevelName(loaded["logging"]["level"])
            else:
                log_level = None
            if "to_file" in loaded["logging"]:
                log_to_file = bool(loaded["logging"]["to_file"])
            else:
                log_to_file = None
            if "rotate" in loaded["logging"]:
                log_rotate = bool(loaded["logging"]["rotate"])
            else:
                log_rotate = None
        else:
            log_level = None
            log_to_file = False
            log_rotate = False
        if "cleanup" in loaded:
            if "auto" in loaded["cleanup"]:
                cleanup_auto = bool(loaded["cleanup"]["auto"])
            else:
                cleanup_auto = False
            if "auto" in loaded["cleanup"]:
                cleanup_keep = int(loaded["cleanup"]["keep"])
            else:
                cleanup_keep = 100
        else:
            cleanup_auto = False
            cleanup_keep = 100
        if "paths" in loaded:
            paths = {}
            for key in loaded["paths"]:
                path = loaded["paths"][key]
                if isinstance(path, list):
                    paths[key] = []
                    for p in path:
                        paths[key].append(PathConfig(p, base=home))
                else:
                    paths[key] = PathConfig(path, base=home)
        else:
            paths = None
        if "repos" in loaded:
            repos = {}
            for key in loaded["repos"]:
                repo = loaded["repos"][key]
                url = repo.get("url", None)
                if url is None:
                    raise RuntimeError("Missing field 'url' in YAML file")
                ref = repo.get("ref", None)
                options = repo.get("options", None)
                repos[key] = RepoConfig(url, ref=ref, options=options)
        else:
            repos = None
        default_framework = None
        default_backends = {}
        if "frameworks" in loaded:
            frameworks = []
            default_framework = loaded["frameworks"].pop("default", None)
            for key in loaded["frameworks"]:
                framework = loaded["frameworks"][key]
                if "enabled" in framework:
                    enabled = bool(framework["enabled"])
                else:
                    enabled = False
                backends = []
                if "backends" in framework:
                    default_backend = framework["backends"].pop("default", None)
                    default_backends[key] = default_backend
                    for key2 in framework["backends"]:
                        backend = framework["backends"][key2]
                        if "enabled" in backend:
                            enabled2 = bool(backend["enabled"])
                        else:
                            enabled2 = True
                        backend_features = []
                        if "features" in backend:
                            for key3 in backend["features"]:
                                supported = bool(backend["features"][key3])
                                backend_features.append(BackendFeatureConfig(key3, backend=key2, supported=supported))
                        backends.append(BackendConfig(key2, enabled=enabled2, features=backend_features))
                framework_features = []
                if "features" in framework:
                    for key2 in framework["features"]:
                        supported = bool(framework["features"][key2])
                        framework_features.append(FrameworkFeatureConfig(key2, framework=key, supported=supported))
                frameworks.append(
                    FrameworkConfig(
                        key,
                        enabled=enabled,
                        backends=backends,
                        features=framework_features,
                    )
                )
        else:
            frameworks = None
        if "frontends" in loaded:
            frontends = []
            for key in loaded["frontends"]:
                frontend = loaded["frontends"][key]

                if "enabled" in frontend:
                    enabled = frontend["enabled"]
                else:
                    enabled = True
                frontend_features = []
                if "features" in frontend:
                    for key2 in frontend["features"]:
                        supported = bool(frontend["features"][key2])
                        frontend_features.append(FrontendFeatureConfig(key2, frontend=key, supported=supported))
                frontends.append(FrontendConfig(key, enabled=enabled, features=frontend_features))
        else:
            frontends = None
        if "platforms" in loaded:
            platforms = []
            for key in loaded["platforms"]:
                platform = loaded["platforms"][key]
                if "enabled" in platform:
                    enabled = platform["enabled"]
                else:
                    enabled = True
                platform_features = []
                if "features" in platform:
                    for key2 in platform["features"]:
                        supported = bool(platform["features"][key2])
                        platform_features.append(PlatformFeatureConfig(key2, platform=key, supported=supported))
                platforms.append(PlatformConfig(key, enabled=enabled, features=platform_features))
        else:
            platforms = None
        if "toolchains" in loaded:
            toolchains = loaded["toolchains"]
        else:
            toolchains = None
        default_target = None
        if "targets" in loaded:
            targets = []
            default_target = loaded["targets"].pop("default", None)
            for key in loaded["targets"]:
                target = loaded["targets"][key]
                if "enabled" in target:
                    enabled = target["enabled"]
                else:
                    enabled = True
                target_features = []
                if "features" in target:
                    for key2 in target["features"]:
                        supported = bool(target["features"][key2])
                        target_features.append(TargetFeatureConfig(key2, target=key, supported=supported))
                targets.append(TargetConfig(key, enabled=enabled, features=target_features))
        else:
            targets = None
        if "vars" in loaded:
            variables = loaded["vars"]
        else:
            variables = None
        if "flags" in loaded:
            default_flags = loaded["flags"]
        else:
            default_flags = None
        defaults = DefaultsConfig(
            log_level=log_level,
            log_to_file=log_to_file,
            log_rotate=log_rotate,
            default_framework=default_framework,
            default_backends=default_backends,
            default_target=default_target,
            cleanup_auto=cleanup_auto,
            cleanup_keep=cleanup_keep,
        )
        env = base(
            home,
            defaults=defaults,
            paths=paths,
            repos=repos,
            frameworks=frameworks,
            frontends=frontends,
            platforms=platforms,
            toolchains=toolchains,
            targets=targets,
            variables=variables,
            default_flags=default_flags,
        )
        return env



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def get_autotuning_defaults():
    return {
        "mode": None,
        "results_file": None,
        "append": None,
        "trials": 10,  # TODO: increase to 100?
        "trials_single": None,
        "early_stopping": None,  # calculate default dynamically
        "num_workers": None,
        "max_parallel": 1,
        "use_rpc": False,
        "timeout": 100,
        "tasks": None,
        "visualize": False,
        "visualize_file": None,
        "visualize_live": False,
    }


def get_autotvm_defaults():
    return {
        "enable": False,
        "tuner": "ga",  # Options: ga,gridsearch,random,xgb,xgb_knob,xgb-rank
    }


def get_autoscheduler_defaults():
    return {
        "enable": False,
        "include_simple_tasks": False,
        "log_estimated_latency": True,
    }


def get_metascheduler_defaults():
    return {
        "enable": False,
    }



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Collection of utilities to manage MLonMCU configs."""
import distutils.util
import ast

from mlonmcu.feature.type import FeatureType
from mlonmcu.logging import get_logger

logger = get_logger()


def remove_config_prefix(config, prefix, skip=None):
    """Iterate over keys in dict and remove given prefix.

    Arguments
    ---------
    config : dict
        The configuration data.
    prefix : str
        The prefix to remove.
    skip : List[str], optional
        A list of keys which should not be altered.

    Returns
    -------
    ret : dict
        The transformed configuration.

    """
    if skip is None:
        skip = []

    def helper(key):
        return key.split(f"{prefix}.")[-1]

    return {helper(key): value for key, value in config.items() if f"{prefix}." in key and key not in skip}


def filter_config(config, prefix, defaults, optionals, required_keys):
    """Filter the global config for a given component prefix.

    Arguments
    ---------
    config : dict
        The configuration data.
    prefix : str
        The prefix for the component.
    defaults : dict
        The default values used if not overwritten by user.
    optionals : dict
        The optional keys for the component.
    required_keys : list
        The required keys for the component.

    Returns
    -------
    cfg : dict
        The filteres configuration.

    Raises
    ------
    AssertionError: If a required key is missing.
    """
    if not isinstance(required_keys, set):
        logger.warning(
            "Deprecated: FEATURES, REQUIRED, OPTIONAL should now be sets, not lists for component %s", prefix
        )
        required_keys = set(required_keys)
    if not isinstance(optionals, set):
        optionals = set(optionals)
    cfg = remove_config_prefix(config, prefix, skip=required_keys | optionals)
    for required in required_keys:
        value = None
        if required in cfg:
            value = cfg[required]
        elif required in config:
            value = config[required]
            cfg[required] = value
        assert value is not None, f"Required config key can not be None: {required}"

    for optional in optionals:
        if optional in cfg:
            continue
        if optional in config:
            value = config[optional]
            cfg[optional] = value
        else:
            cfg[optional] = None

    for key in defaults:
        if key not in cfg:
            cfg[key] = defaults[key]

    for key in cfg:
        if key not in set(defaults.keys()) | required_keys:
            # logger.warn("Component received an unknown config key: %s", key)
            pass

    return cfg


def resolve_required_config(
    required_keys, optional=None, features=None, config=None, cache=None, hints=None, default_flags=None
):  # TODO: add framework, backend, and frontends as well?
    """Utility which iterates over a set of given config keys and
    resolves their values using the passed config and/or cache.

    Parameters
    ----------
    required_keys : List[str]

    features : List[Feature]

    config : dict

    cache : TaskCache
        Optional task cache parsed from the `cache.ini` file in the `deps` directory.

    hints : List[str]
        List of additional flags which can be provided as a hint to lookup a cache config.

    default_flags : dict
        User-provided mapping of cache flags for some cache entries.

    Returns
    -------
    result : dict

    """

    def get_sublists(in_list):
        ret = [[]]
        for i in range(len(in_list) + 1):
            for j in range(i + 1, len(in_list) + 1):
                sub = in_list[i:j]
                ret.append(sub)
        return ret

    hint_combinations = get_sublists(hints if hints else [])

    def get_cache_flags(features, default):
        result = {}
        if default:
            # result.update({key: frozenset(value) for key, value in default.items()})
            result.update(default)
        if features:
            for feature in features:
                if FeatureType.SETUP in type(feature).types():
                    feature.add_required_cache_flags(result)
        return result

    ret = {}
    cache_flags = get_cache_flags(features, default_flags)
    for key in required_keys:
        if config is None or key not in config:
            assert cache is not None, "No dependency cache was provided. Either provide a cache or config."
            if len(cache) == 0:
                raise RuntimeError("The dependency cache is empty! Make sure `to run `mlonmcu` setup first.`")
            flags = cache_flags.get(key, ())
            value = None
            for hint_combination in hint_combinations:
                if (key, tuple(list(flags) + hint_combination)) in cache:
                    value = cache[key, flags]
                    break
            if value is None:
                if len(flags) == 0:
                    raise RuntimeError(
                        f"Dependency cache miss for required key '{key}'. Try re-running `mlonmcu setup`."
                    )
                else:
                    raise RuntimeError(
                        f"Dependency cache miss for required key '{key}' with flags {flags}."
                        + " Try re-running `mlonmcu setup`."
                    )
            ret[key] = value
        else:
            ret[key] = config[key]
    if optional:
        for key in optional:
            if config is None or key not in config:
                assert cache is not None, "No dependency cache was provided. Either provide a cache or config."
                if cache is None or len(cache) == 0:
                    continue
                flags = cache_flags.get(key, ())
                value = None
                for hint_combination in hint_combinations:
                    if (key, tuple(list(flags) + hint_combination)) in cache:
                        value = cache[key, flags]
                        break
                if value is None:
                    continue
                ret[key] = value
            else:
                ret[key] = config[key]

    return ret


def str2bool(value, allow_none=False):
    if value is None:
        assert allow_none, "str2bool received None value while allow_none=False"
        return value
    if isinstance(value, (int, bool)):
        return bool(value)
    assert isinstance(value, str)
    return bool(distutils.util.strtobool(value))


def str2dict(value, allow_none=False):
    if value is None:
        assert allow_none, "str2dict received None value while allow_none=False"
        return value
    if isinstance(value, dict):
        return value
    assert isinstance(value, str)
    # TODO: parse key=value,key2=value2 via regex
    return dict(ast.literal_eval(value))


def str2list(value, allow_none=False):
    if value is None:
        assert allow_none, "str2list received None value while allow_none=False"
        return value
    if isinstance(value, (list, set, tuple)):
        return list(value)
    assert isinstance(value, str)
    if value.startswith("["):
        return list(ast.literal_eval(value))
    return value.split(",")



#!/usr/bin/env python
#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""ELF File Tool"""

# import os
# import sys
import csv
import argparse
from elftools.elf import elffile

from mlonmcu.logging import get_logger

logger = get_logger()


"""
Script to gather metrics on static ROM and RAM usage.

Heavility inspired by get_metrics.py found in the ETISS repository
"""


def parseElf(inFile):
    """Extract static memory usage details from ELF file by mapping each segment."""
    # TODO: check if this is generic anough for multiple platforms (riscv, arm, x86)
    # TODO: comare results with `riscv32-unknown-elf-size`
    m = {}
    m["rom_rodata"] = 0
    m["rom_code"] = 0
    m["rom_misc"] = 0
    m["ram_data"] = 0
    m["ram_zdata"] = 0

    ignoreSections = [
        "",
        ".stack",
        ".comment",
        ".riscv.attributes",
        ".strtab",
        ".stabstr",
        ".shstrtab",
        ".symtab",
        ".eh_frame",
        ".stab",
        ".heap",  # ?
        # The following are x86 only:
        ".interp",
        ".dynsym",
        ".dynstr",
        ".dynamic",
        ".got",
        ".data.rel.ro",
        # Espressif
        ".flash.appdesc",
        ".iram0.text_end",  # ?
        # QEMU
        ".htif",
        # Zephyr
        ".mcuboot_header",
        ".metadata",
        "ctors",
        "initlevel",
        "devices",
        "device_handles",
        "sw_isr_table",
        "device_states",
        ".mcuboot_header",
        ".metadata",
        "ctors",
        "initlevel",
        "devices",
        "device_handles",
        "sw_isr_table",
        "device_states",
        ".xt.prop",
        ".xt.lit",
        "k_heap_area",
        "datas",
        # Pulp
        ".data_tiny_fc",
        ".data_tiny_l1",
        ".l1cluster_g",
        ".heap_l2_shared",
        ".Pulp_Chip.Info",
        # ARM (corstone300)
        ".ddr",
        # cv32e40p
        ".debugger_stack",
        # ara
        ".l2",
        # vicuna (ram)
        ".user_align",
    ]
    ignorePrefixes = [
        ".gcc_except",
        ".sdata2",
        ".debug_",
        # ARM only:
        ".ARM",
        # The following are x86 only:
        ".note",
        ".gnu",
        ".rela",
        ".plt",
    ]
    ignoreSuffixes = [
        ".table",
        "dummy",
        "heap_start",
        "rom_start",
        ".info",
    ]

    with open(inFile, "rb") as f:
        e = elffile.ELFFile(f)

        for s in e.iter_sections():
            if s.name.startswith(".text") or s.name.endswith(".text") or s.name == "text":
                m["rom_code"] += s.data_size
            elif s.name.startswith(".srodata"):
                m["rom_rodata"] += s.data_size
            elif s.name.startswith(".sdata"):
                m["ram_data"] += s.data_size
            elif s.name.endswith(".rodata") or s.name == "rodata":
                m["rom_rodata"] += s.data_size
            elif s.name in [
                ".vectors",
                "iram0.vectors",
                ".iram0.vectors",
                ".init_array",
                ".fini_array",
                ".fini",
                ".init",
                ".eh_frame",
                ".eh_frame_hdr",
            ]:
                m["rom_misc"] += s.data_size
            elif s.name.endswith(".data"):
                m["ram_data"] += s.data_size
            elif (
                s.name == ".bss"
                or s.name == "bss"
                or s.name == ".sbss"
                or s.name == ".shbss"
                or s.name == ".bss.noinit"
                or s.name.endswith(".bss")
                or s.name.startswith(".bss")
                or s.name.startswith(".sbss")
                or s.name == "noinit"
            ):
                m["ram_zdata"] += s.data_size
            elif s.name in ignoreSections:
                pass
            elif any(s.name.startswith(prefix) for prefix in ignorePrefixes):
                pass
            elif any(s.name.endswith(suffix) for suffix in ignoreSuffixes):
                pass
            elif s.data_size == 0:
                pass  # No warning for empty sections
            else:
                logger.warning("ignored: %s / size: %d", s.name, s.data_size)

    return m


def printSz(sz, unknown_msg=""):
    """Helper function for printing file sizes."""
    if sz is None:
        return f"unknown [{unknown_msg}]" if unknown_msg else "unknown"
    import humanize

    return humanize.naturalsize(sz) + " (" + hex(sz) + ")"


def parse_cmdline():
    """Cmdline interface definition."""
    parser = argparse.ArgumentParser()
    parser.add_argument("elf", metavar="ELF", type=str, nargs=1, help="The target ELF file")
    parser.add_argument(
        "--out",
        "-o",
        metavar="FILE",
        type=str,
        default="",
        help="""Output CSV file (default: -)""",
    )
    args = parser.parse_args()

    elfFile = args.elf[0]
    csvFile = args.out

    return elfFile, csvFile


def get_results(elfFile):
    """Converts and returns collected data."""
    staticSizes = parseElf(elfFile)

    romSize = sum([size for key, size in staticSizes.items() if key.startswith("rom_")])
    ramSize = sum([size for key, size in staticSizes.items() if key.startswith("ram_")])

    results = {
        "rom": romSize,
        "rom_rodata": staticSizes["rom_rodata"],
        "rom_code": staticSizes["rom_code"],
        "rom_misc": staticSizes["rom_misc"],
        "ram": ramSize,
        "ram_data": staticSizes["ram_data"],
        "ram_zdata": staticSizes["ram_zdata"],
    }
    return results


def print_results(results):
    """Displaying a fancy overview."""
    print("=== Results ===")
    print("ROM usage:        " + printSz(results["rom"]))
    print("  read-only data: " + printSz(results["rom_rodata"]))
    print("  code:           " + printSz(results["rom_code"]))
    print("  other required: " + printSz(results["rom_misc"]))
    print("RAM usage:        " + printSz(results["ram"]))
    print("  data:           " + printSz(results["ram_data"]))
    print("  zero-init data: " + printSz(results["ram_zdata"]))


def write_csv(filename, results):
    """Utility for writing a CSV file."""
    # Write metrics to file
    if filename:
        with open(filename, "w", encoding="utf-8") as f:
            writer = csv.DictWriter(f, fieldnames=results.keys())
            writer.writeheader()
            writer.writerow(results)


def main():
    """Main entry point for command line usage."""
    elfFile, csvFile = parse_cmdline()

    results = get_results(elfFile)

    print_results(results)

    if csvFile:
        write_csv(csvFile, results)


if __name__ == "__main__":
    main()
